"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""

import copy
import re
import six
import warnings

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


@types.validator_msg("Is x any type")
def _is_anytype(_):
    return True


_Anytype = types.BasicType("_anytype", _is_anytype)


@configsuite.validator_msg("AllowNone can only be used for BasicType")
def _check_allownone_type(schema_level):
    if MK.AllowNone in schema_level:
        return isinstance(schema_level[MK.Type], types.BasicType)
    return True


@configsuite.validator_msg("Required can only be used for BasicType")
def _check_required_type(schema_level):
    if MK.Required in schema_level:
        return isinstance(schema_level[MK.Type], types.BasicType)
    return True


@configsuite.validator_msg(
    "A type is not required only if it allows None or have a non-None default"
)
def _check_allownone_required(schema_level):
    if not isinstance(schema_level[MK.Type], types.BasicType):
        return True

    is_required = schema_level.get(MK.Required, True)
    allow_none = schema_level.get(MK.AllowNone, False)
    has_non_none_default = schema_level.get(MK.Default) is not None

    return is_required != allow_none or has_non_none_default


@configsuite.validator_msg("Default can only be used for BasicType")
def _check_default_type(schema_level):
    if MK.Default in schema_level:
        return isinstance(schema_level[MK.Type], types.BasicType)
    return True


@configsuite.validator_msg("Required can not have Default")
def _check_required_not_default(schema_level):
    if schema_level.get(MK.Required, True):
        return MK.Default not in schema_level
    return True


_SCHEMA_LEVEL_DEFAULTS = {
    MK.Required: True,
    MK.AllowNone: False,
    MK.Description: "",
    MK.ElementValidators: (),
    MK.ContextValidators: (),
}


META_SCHEMA = {
    MK.Type: types.NamedDict,
    MK.ElementValidators: (
        _check_allownone_type,
        _check_required_type,
        _check_default_type,
        _check_allownone_required,
        _check_required_not_default,
    ),
    MK.Content: {
        MK.Type: {MK.Type: types.Type},
        MK.Required: {
            MK.Type: types.Bool,
            MK.Required: False,
            MK.Default: _SCHEMA_LEVEL_DEFAULTS[MK.Required],
        },
        MK.AllowNone: {
            MK.Type: types.Bool,
            MK.Required: False,
            MK.Default: _SCHEMA_LEVEL_DEFAULTS[MK.AllowNone],
        },
        MK.Description: {
            MK.Type: types.String,
            MK.Required: False,
            MK.Default: _SCHEMA_LEVEL_DEFAULTS[MK.Description],
        },
        MK.Default: {MK.Type: _Anytype, MK.Required: False, MK.AllowNone: True},
        MK.ElementValidators: {
            MK.Type: types.List,
            MK.Content: {MK.Item: {MK.Type: types.Callable}},
        },
        MK.ContextValidators: {
            MK.Type: types.List,
            MK.Content: {MK.Item: {MK.Type: types.Callable}},
        },
        MK.Transformation: {
            MK.Type: types.Callable,
            MK.Required: False,
            MK.AllowNone: True,
        },
        MK.ContextTransformation: {
            MK.Type: types.Callable,
            MK.Required: False,
            MK.AllowNone: True,
        },
        MK.LayerTransformation: {
            MK.Type: types.Callable,
            MK.Required: False,
            MK.AllowNone: True,
        },
    },
}


def _build_meta_schema(deduce_required):
    meta_schema = copy.deepcopy(META_SCHEMA)
    if deduce_required:
        meta_schema[MK.ElementValidators] = (
            _check_allownone_type,
            _check_required_type,
            _check_default_type,
        )

    return meta_schema


_REQUIRED_DEPRECATION_MSG = (
    "Specifying whether elements are required directly is deprecated. "
    "Please remove them from your schema to adopt to future behaviour. "
    "See the documentation for more details."
)


_EXPLICIT_REQUIRED_DEPRECATION_MSG = (
    "Specifying whether elements are required directly is deprecated. Use "
    "`ConfigSuite(..., deduce_required=True)` to adopt to future behaviour. "
    "See the documentation for more details."
)


def assert_valid_schema(
    schema, allow_default=False, validate_named_keys=True, deduce_required=False
):
    with warnings.catch_warnings(record=True) as warnings_manager:
        _assert_valid_schema(
            schema,
            allow_default=allow_default,
            validate_named_keys=validate_named_keys,
            deduce_required=deduce_required,
        )

    if not deduce_required:
        warnings.warn(
            _EXPLICIT_REQUIRED_DEPRECATION_MSG, DeprecationWarning, stacklevel=3,
        )
    elif any(_REQUIRED_DEPRECATION_MSG == str(w.message) for w in warnings_manager):
        warnings.warn(
            _REQUIRED_DEPRECATION_MSG, DeprecationWarning, stacklevel=3,
        )


def _assert_valid_schema(schema, allow_default, validate_named_keys, deduce_required):
    _assert_valid_schema_level(schema, allow_default, deduce_required)

    level_type = schema[MK.Type]
    if isinstance(level_type, types.BasicType):
        return
    elif level_type == types.NamedDict:
        _assert_valid_named_dict_schema(schema, validate_named_keys, deduce_required)
    elif level_type == types.List:
        _assert_valid_list_schema(schema, validate_named_keys, deduce_required)
    elif level_type == types.Dict:
        _assert_valid_dict_schema(schema, validate_named_keys, deduce_required)
    else:
        raise TypeError("Unknown base container: {}".format(schema))


def _build_level_schema(schema):
    schema = copy.deepcopy(schema)

    # Make schema into a pure level schema
    if MK.Content in schema:
        schema.pop(MK.Content)

    level_schema = copy.deepcopy(_SCHEMA_LEVEL_DEFAULTS)

    # Discard ignore from default if not in level schema
    if MK.Required not in schema:
        level_schema.pop(MK.Required)

    # Discard basic type defaults for non-basic types
    if not isinstance(schema[MK.Type], types.BasicType):
        for basic_key in (MK.Required, MK.Default, MK.AllowNone):
            if basic_key in level_schema:
                level_schema.pop(basic_key)

    level_schema.update(schema)
    return level_schema


def _assert_valid_schema_level(schema, allow_default, deduce_required):
    schema = _build_level_schema(schema)

    if MK.Required in schema:
        warnings.warn(_REQUIRED_DEPRECATION_MSG, DeprecationWarning)

    if MK.Default in schema and not allow_default:
        fmt = "Default value is only allowed for contents in NamedDict"
        raise ValueError(fmt)

    level_validator = configsuite.Validator(_build_meta_schema(deduce_required))
    result = level_validator.validate(schema)

    if not result.valid:
        for error in result.errors:

            def gen_err_msg(error):
                return "{} at {}".format(error.msg, error.key_path)

            if isinstance(error, configsuite.MissingKeyError):
                raise KeyError(gen_err_msg(error))
            elif isinstance(error, configsuite.UnknownKeyError):
                raise KeyError(gen_err_msg(error))
            elif isinstance(error, configsuite.InvalidTypeError):
                raise TypeError(gen_err_msg(error))
            elif isinstance(error, configsuite.InvalidValueError):
                raise ValueError(gen_err_msg(error))
            else:
                raise AssertionError(
                    "Internal error: Unknown validation error: {}".format(error)
                )


def _assert_dict_key(key):
    if not isinstance(key, six.string_types):
        raise KeyError(
            "Expected all {} keys to be strings, found: {}"
            "".format(types.NamedDict.name, type(key))
        )

    key_regex = "^[a-zA-Z_][a-zA-Z0-9_]*$"
    if not re.match(key_regex, key):
        raise KeyError(
            'Expected all {} keys to match: "{}", found: {}'
            "".format(types.NamedDict.name, key_regex, key)
        )


def _assert_valid_named_dict_schema(schema, validate_keys, deduce_required):
    if MK.Content not in schema:
        err_msg = "{} schema has no {}: {}".format(
            types.NamedDict.name, MK.Content, schema
        )
        raise KeyError(err_msg)

    content = schema[MK.Content]
    if not isinstance(content, dict):
        err_msg = "Expected {} to be a dict, was {}".format(MK.Content, type(content))
        raise ValueError(err_msg)

    if validate_keys:
        for key in content.keys():
            _assert_dict_key(key)

    for value in content.values():
        _assert_valid_schema(value, True, validate_keys, deduce_required)


def _assert_valid_list_schema(schema, validate_named_keys, deduce_required):
    if MK.Content not in schema:
        err_msg = "{} schema has no {}: {}".format(types.List.name, MK.Content, schema)
        raise KeyError(err_msg)

    content = schema[MK.Content]
    if not isinstance(content, dict):
        err_msg = "Expected {} to be a dict, was {}".format(MK.Content, type(content))
        raise ValueError(err_msg)

    if tuple(content.keys()) != (MK.Item,):
        err_msg = (
            "Expected {} of a {} to contain exactly one key. "
            ", namely {}, was {}."
            "".format(MK.Content, types.List.name, MK.Item, content.keys())
        )
        raise KeyError(err_msg)

    _assert_valid_schema(content[MK.Item], False, validate_named_keys, deduce_required)


def _assert_valid_dict_schema(schema, validate_named_keys, deduce_required):
    if MK.Content not in schema:
        err_msg = "{} schema has no {}: {}".format(types.Dict.name, MK.Content, schema)
        raise KeyError(err_msg)

    content = schema[MK.Content]
    if not isinstance(content, dict):
        err_msg = "Expected {} to be a dict, was {}".format(MK.Content, type(content))
        raise ValueError(err_msg)

    expected_keys = (MK.Key, MK.Value)
    if len(content.keys()) != len(expected_keys) or set(content.keys()) != set(
        expected_keys
    ):
        err_msg = "Expected {} of a {} to contain the keys {}, was {}." "".format(
            MK.Content, types.Dict.name, (MK.Key, MK.Value), content.keys()
        )
        raise KeyError(err_msg)

    _assert_valid_schema(content[MK.Key], False, validate_named_keys, deduce_required)
    _assert_valid_schema(content[MK.Value], False, validate_named_keys, deduce_required)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""
import os

import contextlib
import logging
import shutil
import tempfile


def relpath(*path):
    """Make a path relative to the project root folder"""
    project_root_folder = os.path.dirname(os.path.dirname(__file__))
    return os.path.join(project_root_folder, *path)


def tmpdir(path=None, teardown=True):
    """ Decorator based on the  `tmp` context """

    def decorator(f):
        def wrapper(*args, **kwargs):
            with tmp(path=path, teardown=teardown):
                return f(*args, **kwargs)

        return wrapper

    return decorator


@contextlib.contextmanager
def tmp(path=None, teardown=True):
    """Create and go into tmp directory, returns the path.

    This function creates a temporary directory and enters that directory.  The
    returned object is the path to the created directory.

    If @path is not specified, we create an empty directory, otherwise, it must
    be a path to an existing directory.  In that case, the directory will be
    copied into the temporary directory.

    If @teardown is True (defaults to True), the directory is (attempted)
    deleted after context, otherwise it is kept as is.

    """
    cwd = os.getcwd()
    fname = tempfile.NamedTemporaryFile().name

    if path:
        if not os.path.isdir(path):
            logging.debug("tmp:raise no such path")
            raise IOError("No such directory: %s" % path)
        shutil.copytree(path, fname)
    else:
        # no path to copy, create empty dir
        os.mkdir(fname)

    os.chdir(fname)

    yield fname  # give control to caller scope

    os.chdir(cwd)

    if teardown:
        try:
            shutil.rmtree(fname)
        except OSError as oserr:
            logging.debug("tmp:rmtree failed %s (%s)", fname, oserr)
            shutil.rmtree(fname, ignore_errors=True)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import collections
import numbers
import datetime
import six


class BooleanResult(object):
    """BooleanResult is a wrapper around a bool that also has a .msg attribute.

    The message is an explanation of the boolean value.
    """

    def __init__(self, value, msg, indata):
        if isinstance(value, BooleanResult):
            value = bool(value)
        if not isinstance(value, bool):
            raise TypeError(
                "Expected the type of value to be bool, was {}".format(type(value))
            )

        self._value = value
        self._msg = str(msg)
        self._input = str(indata)

    def __nonzero__(self):
        return self._value is True

    def __bool__(self):
        return self._value is True

    def __and__(self, other):
        return bool(self) and bool(other)

    @property
    def msg(self):
        msg_fmt = "{} is {} on input '{}'"
        return msg_fmt.format(self._msg, "true" if self else "false", self._input)

    def __repr__(self):
        fmt = "BooleanResult({}, {}, {})"
        return fmt.format(bool(self), self._msg, self._input)


def transformation_msg(msg):
    """Used to decorate a transformation function with a msg.
    """

    def real_decorator(function):
        class Wrapper(object):
            def __init__(self, function, msg):
                self._function = function
                self._msg = msg

            @property
            def msg(self):
                return self._msg

            def __call__(self, *args, **kwargs):
                return self._function(*args, **kwargs)

        return Wrapper(function, msg)

    return real_decorator


def validator_msg(msg):
    """Validator decorator wraps return value in a message container.

    Usage:

        @validator_msg('assert len(x) <= 2')
        def validate_size(x):
            return len(x) <= 2

    Now, if `validate_size` returns a falsy value `ret`, for instance if
    provided with range(4), we will have
        `ret.msg = 'assert len(x) <= 2 is false on input [0, 1, 2, 3]`.

    On the other hand, if `validate_size` returns a true value `ret`, for
    instance if provided with [0, 1], we will have
        `ret.msg = 'assert len(x) <= 2 is true on input [0, 1]`.
    """

    def real_decorator(function):
        def _build_argument_str(*args, **kwargs):
            elems = [str(arg) for arg in args]
            elems += [
                "{}={}".format(str(key), str(value)) for key, value in kwargs.items()
            ]

            return ", ".join(elems)

        class Wrapper(object):
            def __init__(self, function, msg):
                self._function = function
                self._msg = msg

            @property
            def msg(self):
                return self._msg

            def __call__(self, *args, **kwargs):
                res = self._function(*args, **kwargs)
                argument_str = _build_argument_str(*args, **kwargs)
                return BooleanResult(res, self._msg, argument_str)

        return Wrapper(function, msg)

    return real_decorator


BasicType = collections.namedtuple("Type", ["name", "validate"])
Collection = collections.namedtuple("Type", ["name", "validate", "create_empty"])


def _type_eq(self, other):
    return self.name == other.name


def _type_neq(self, other):
    return not _type_eq(self, other)


Collection.__eq__ = _type_eq
Collection.__neq__ = lambda self, other: not _type_neq(self, other)


@validator_msg("Is x a dictionary")
def _is_pydict(x):
    return isinstance(x, dict)


@validator_msg("Is x a list")
def _is_list(x):
    return isinstance(x, (list, tuple))


@validator_msg("Is x a string")
def _is_string(x):
    return isinstance(x, six.string_types)


@validator_msg("Is x an integer")
def _is_integer(x):
    return isinstance(x, int)


@validator_msg("Is x a number")
def _is_number(x):
    return isinstance(x, numbers.Number)


@validator_msg("Is x a bool")
def _is_bool(x):
    return isinstance(x, bool)


@validator_msg("Is x a valid ISO8601(YYYY-mm-dd) date format")
def _is_date(x):
    return isinstance(x, datetime.date)


@validator_msg("Is x a datetime")
def _is_datetime(x):
    return isinstance(x, datetime.datetime)


NamedDict = Collection("named_dict", _is_pydict, dict)
Dict = Collection("dict", _is_pydict, dict)
List = Collection("list", _is_list, tuple)
String = BasicType("string", _is_string)
Integer = BasicType("integer", _is_integer)
Number = BasicType("number", _is_number)
Bool = BasicType("bool", _is_bool)
Date = BasicType("date", _is_date)
DateTime = BasicType("datetime", _is_datetime)

# Meta types


@validator_msg("Is x a type")
def _is_type(x):
    return isinstance(x, (BasicType, Collection))


Type = BasicType("type", _is_type)
Callable = BasicType("callable", validator_msg("Is x callable")(callable))



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import datetime
import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


def build_schema():
    @configsuite.validator_msg("Score should be non-negative")
    def _non_negative_score(elem):
        return elem > 0

    return {
        MK.Type: types.NamedDict,
        MK.Description: (
            "The PetMeet configuration is to give an overview of you and your "
            "pet to better be able to find a matching pet owner."
        ),
        MK.Content: {
            "name": {MK.Type: types.String, MK.Description: "Your full name."},
            "pet": {
                MK.Type: types.NamedDict,
                MK.Description: "Information about your pet.",
                MK.Content: {
                    "name": {MK.Type: types.String, MK.Description: "Name of the pet."},
                    "timestamp": {
                        MK.Type: types.DateTime,
                        MK.Description: "Record entry date.",
                    },
                    "favourite_food": {
                        MK.Type: types.String,
                        MK.Description: "Favourite food of the pet.",
                    },
                    "weight": {
                        MK.Type: types.Number,
                        MK.Description: "Weight of pet in tons.",
                    },
                    "nkids": {
                        MK.Type: types.Integer,
                        MK.Description: "Number of kids.",
                    },
                    "likeable": {
                        MK.Type: types.Bool,
                        MK.Description: "Is the pet likeable?",
                    },
                    "birthday": {
                        MK.Type: types.Date,
                        MK.Description: "Birthday of the pet.",
                    },
                },
            },
            "playgrounds": {
                MK.Type: types.List,
                MK.Description: "List of all playgrounds you take your pet to.",
                MK.Content: {
                    MK.Item: {
                        MK.Type: types.NamedDict,
                        MK.Description: "Information about a playground.",
                        MK.Content: {
                            "name": {
                                MK.Type: types.String,
                                MK.Description: "Name of the playground.",
                            },
                            "score": {
                                MK.Type: types.Integer,
                                MK.Description: "Your score for the playground.",
                            },
                        },
                    }
                },
            },
            "veterinary_scores": {
                MK.Type: types.Dict,
                MK.Description: (
                    "List of all veterinaries you have taken your pet to."
                ),
                MK.Content: {
                    MK.Key: {
                        MK.Type: types.String,
                        MK.Description: "Name of the veterinary.",
                    },
                    MK.Value: {
                        MK.ElementValidators: (_non_negative_score,),
                        MK.Type: types.Number,
                        MK.Description: "Your score of the vet.",
                    },
                },
            },
        },
    }


def build_config():
    return {
        "name": "Markus",
        "pet": {
            "name": "Donkey Kong",
            "timestamp": datetime.datetime(2018, 5, 6, 12, 58, 22),
            "favourite_food": "bananas",
            "weight": 1.9,
            "nkids": 8,
            "likeable": True,
            "birthday": datetime.date(2018, 1, 1),
        },
        "playgrounds": [],
        "veterinary_scores": {"Dr. Dolittle": 10},
    }


def build_docs():
    return """The PetMeet configuration is to give an overview of you and your \
pet to better be able to find a matching pet owner.

**name*:**
    Your full name.

    :type: string

**pet*:**
    Information about your pet.

    **name*:**
        Name of the pet.

        :type: string

    **timestamp*:**
        Record entry date.

        :type: datetime

    **favourite_food*:**
        Favourite food of the pet.

        :type: string

    **weight*:**
        Weight of pet in tons.

        :type: number

    **nkids*:**
        Number of kids.

        :type: integer

    **likeable*:**
        Is the pet likeable?

        :type: bool

    **birthday*:**
        Birthday of the pet.

        :type: date

**playgrounds*:**
    List of all playgrounds you take your pet to.

    **<list_item>:**
            Information about a playground.

            **name*:**
                Name of the playground.

                :type: string

            **score*:**
                Your score for the playground.

                :type: integer

**veterinary_scores*:**
    List of all veterinaries you have taken your pet to.

    **<key>:**
            Name of the veterinary.

            :type: string

    **<value>:**
            Your score of the vet.

            :requirement: Score should be non-negative

            :type: number"""



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import collections
import configsuite
from configsuite import MetaKeys as MK


ValidationResult = collections.namedtuple("ValidationResult", ("valid", "errors"))


class Validator(object):
    def __init__(
        self, schema, stop_condition=(lambda schema: False), apply_validators=True
    ):
        self._schema = schema
        self._errors = None
        self._key_stack = None
        self._context = None
        self._stop_condition = stop_condition
        self._apply_validators = apply_validators

    def validate(self, config, context=None):
        self._errors = []
        self._key_stack = _KeyStack()
        self._context = context
        valid = self._validate(config, self._schema)
        return ValidationResult(valid=valid, errors=tuple(self._errors))

    def _validate(self, config, schema):
        if self._stop_condition(schema):
            return True

        data_type = schema[MK.Type]
        allow_none = schema.get(MK.AllowNone, False)

        if allow_none and config is None:
            valid = True
        else:
            valid = data_type.validate(config)

        if not valid:
            self._add_invalid_type_error(valid.msg)
        elif isinstance(data_type, configsuite.BasicType):
            pass
        elif data_type == configsuite.types.NamedDict:
            valid &= self._validate_named_dict(config, schema[MK.Content])
        elif data_type == configsuite.types.List:
            valid &= self._validate_list(config, schema[MK.Content])
        elif data_type == configsuite.types.Dict:
            valid &= self._validate_dict(config, schema[MK.Content])
        else:
            msg = "Unknown type {} while validating"
            raise TypeError(msg.format(data_type))

        if self._apply_validators and valid:
            valid &= self._element_validation(config, schema)

        if self._apply_validators and valid:
            valid &= self._context_validation(config, schema)

        return bool(valid)

    def _element_validation(self, config, schema):
        elem_vals = schema.get(MK.ElementValidators, ())

        valid = True
        for val in elem_vals:
            res = val(config)
            if not res:
                valid = False
                self._add_invalid_value_error(res.msg)

        return valid

    def _context_validation(self, config, schema):
        context_validators = schema.get(MK.ContextValidators, ())

        valid = True
        for validator in context_validators:
            res = validator(config, self._context)
            if not res:
                valid = False
                self._add_invalid_value_error(res.msg)

        return valid

    def _identify_unknown_dict_keys(self, config, content_schema):
        unknown_keys = set(config.keys()) - set(content_schema.keys())
        for key in unknown_keys:
            msg_fmt = "Unknown key: {}"
            self._add_unknown_key_error(msg_fmt.format(key))
        return len(unknown_keys) == 0

    def _identify_missing_dict_keys(self, config, content_schema):
        optional_keys = []
        for key in content_schema:
            deduced_required = not (
                content_schema[key].get(MK.AllowNone, False)
                or content_schema[key].get(MK.Default, None) is not None
            )

            if not deduced_required:
                optional_keys.append(key)

        missing_keys = (
            set(content_schema.keys()) - set(config.keys()) - set(optional_keys)
        )

        for key in missing_keys:
            msg_fmt = "Missing key: {}"
            self._add_missing_key_error(msg_fmt.format(key))
        return len(missing_keys) == 0

    def _validate_named_dict(self, config, content_schema):
        valid = True
        valid &= self._identify_unknown_dict_keys(config, content_schema)
        valid &= self._identify_missing_dict_keys(config, content_schema)

        valid_keys = set(config.keys()).intersection(set(content_schema.keys()))
        for key in valid_keys:
            self._key_stack.append(key)
            valid &= self._validate(config[key], content_schema[key])
            self._key_stack.pop()

        return valid

    def _validate_list(self, config, schema):
        item_schema = schema[MK.Item]

        valid = True
        for idx, config_item in enumerate(config):
            self._key_stack.append(idx)
            valid &= self._validate(config_item, item_schema)
            self._key_stack.pop()

        return valid

    def _validate_dict(self, config, content_schema):
        key_schema = content_schema[MK.Key]
        value_schema = content_schema[MK.Value]

        valid = True
        for key, value in config.items():
            self._key_stack.append(key)
            valid &= self._validate(key, key_schema)
            valid &= self._validate(value, value_schema)
            self._key_stack.pop()

        return valid

    def _add_invalid_type_error(self, msg):
        self._add_error(msg, configsuite.InvalidTypeError)

    def _add_unknown_key_error(self, msg):
        self._add_error(msg, configsuite.UnknownKeyError)

    def _add_missing_key_error(self, msg):
        self._add_error(msg, configsuite.MissingKeyError)

    def _add_invalid_value_error(self, msg):
        self._add_error(msg, configsuite.InvalidValueError)

    def _add_error(self, msg, ErrorType):
        err = ErrorType(msg, self._key_stack.keys())
        self._errors.append(err)


class _KeyStack(object):
    def __init__(self):
        self._stack = []

    def append(self, key):
        self._stack.append(key)

    def pop(self):
        self._stack.pop()

    def keys(self):
        return tuple(self._stack)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""

import datetime
import numbers
import unittest
import six

import configsuite
from configsuite import MetaKeys as MK

from . import data


class TestTypes(unittest.TestCase):
    def test_name_pet_accepted(self):
        raw_config = data.pets.build_config()
        config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

        self.assertTrue(config_suite.valid)

        config = config_suite.snapshot
        self.assertEqual(raw_config["name"], config.name)
        self.assertEqual(raw_config["pet"]["name"], config.pet.name)
        self.assertEqual(raw_config["pet"]["favourite_food"], config.pet.favourite_food)

    def test_invalid_string(self):
        for name in [14, None, [], (), {}]:
            raw_config = data.pets.build_config()
            raw_config["name"] = name
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            self.assertFalse(config_suite.valid)

            self.assertEqual(1, len(config_suite.errors))
            err = config_suite.errors[0]
            self.assertIsInstance(err, configsuite.InvalidTypeError)
            self.assertEqual(("name",), err.key_path)

            config = config_suite.snapshot
            self.assertEqual(raw_config["name"], config.name)
            self.assertEqual(raw_config["pet"]["name"], config.pet.name)
            self.assertEqual(
                raw_config["pet"]["favourite_food"], config.pet.favourite_food
            )

    def test_invalid_base_dict(self):
        for raw_config in [14, None, [{"name": "Markus"}], ()]:
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())
            self.assertFalse(config_suite.valid)

            self.assertEqual(1, len(config_suite.errors))
            err = config_suite.errors[0]
            self.assertIsInstance(err, configsuite.InvalidTypeError)
            self.assertEqual((), err.key_path)

    def test_invalid_dict(self):
        for pet in [14, None, [{"name": "Markus"}], ()]:
            raw_config = data.pets.build_config()
            raw_config["pet"] = pet
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())
            self.assertFalse(config_suite.valid)

            self.assertEqual(1, len(config_suite.errors))
            err = config_suite.errors[0]
            self.assertIsInstance(err, configsuite.InvalidTypeError)
            self.assertEqual(("pet",), err.key_path)

    def test_string(self):
        for strval in ["fdnsjk", "", "str4thewin", u"ustr", True, [], 1, 1.2, {}, None]:
            raw_config = data.pets.build_config()
            raw_config["name"] = strval
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            if isinstance(strval, six.string_types):
                self.assertTrue(config_suite.valid)
                self.assertEqual(strval, config_suite.snapshot.name)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(("name",), err.key_path)

    def test_int(self):
        for intval in ["fdnsjk", "", False, 0, -10, [], 1, 1.2, {}, None]:
            raw_config = data.pets.build_config()
            raw_config["pet"]["nkids"] = intval
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            if isinstance(intval, int):
                self.assertTrue(config_suite.valid)
                self.assertEqual(intval, config_suite.snapshot.pet.nkids)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(("pet", "nkids"), err.key_path)

    def test_number(self):
        for numval in [False, True, "fdnsjk", "", 0, -10, [], 1, 1.2, {}, None]:
            raw_config = data.pets.build_config()
            raw_config["pet"]["weight"] = numval
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            if isinstance(numval, numbers.Number):
                self.assertTrue(config_suite.valid)
                self.assertEqual(numval, config_suite.snapshot.pet.weight)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(("pet", "weight"), err.key_path)

    def test_bool(self):
        for boolval in [False, True, "fdnsjk", "", 0, -10, [], 1, 1.2, {}, None]:
            raw_config = data.pets.build_config()
            raw_config["pet"]["likeable"] = boolval
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            if isinstance(boolval, bool):
                self.assertTrue(config_suite.valid)
                self.assertEqual(boolval, config_suite.snapshot.pet.likeable)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(("pet", "likeable"), err.key_path)

    def test_list(self):
        playgrounds = [{"name": "superfun", "score": 1}, {"name": "Hell", "score": 99}]

        for end_idx, _ in enumerate(playgrounds):
            raw_config = data.pets.build_config()
            raw_config["playgrounds"] = playgrounds[:end_idx]
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            self.assertTrue(config_suite.valid, "Errors: %r" % (config_suite.errors,))
            config = config_suite.snapshot
            self.assertEqual(end_idx, len(config.playgrounds))

            for idx, plgr in enumerate(playgrounds[:end_idx]):
                self.assertEqual(plgr["name"], config.playgrounds[idx].name)
                self.assertEqual(plgr["score"], config.playgrounds[idx].score)

    def test_containers_in_named_dicts(self):
        schema = {
            MK.Type: configsuite.types.NamedDict,
            MK.Content: {
                "list": {
                    MK.Type: configsuite.types.List,
                    MK.Content: {MK.Item: {MK.Type: configsuite.types.Number}},
                },
                "dict": {
                    MK.Type: configsuite.types.Dict,
                    MK.Content: {
                        MK.Key: {MK.Type: configsuite.types.String},
                        MK.Value: {MK.Type: configsuite.types.Number},
                    },
                },
                "named_dict": {
                    MK.Type: configsuite.types.NamedDict,
                    MK.Content: {
                        "a": {MK.Type: configsuite.types.Number, MK.Default: 0}
                    },
                },
            },
        }

        suite = configsuite.ConfigSuite({}, schema, deduce_required=True)
        self.assertTrue(suite.valid, suite.errors)
        self.assertEqual((), suite.snapshot.list)
        self.assertEqual((), suite.snapshot.dict)
        self.assertEqual(0, suite.snapshot.named_dict.a)

    def test_list_errors(self):
        raw_config = data.pets.build_config()
        raw_config["playgrounds"] = [{"name": "faulty grounds"}]
        config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

        self.assertFalse(config_suite.valid)
        self.assertEqual(1, len(config_suite.errors))
        err = config_suite.errors[0]
        self.assertIsInstance(err, configsuite.MissingKeyError)
        self.assertEqual(("playgrounds", 0), err.key_path)

    def test_list_cannot_spec_required(self):
        schema = {
            MK.Type: configsuite.types.List,
            MK.Content: {MK.Item: {MK.Type: configsuite.types.Integer}},
        }
        config_suite = configsuite.ConfigSuite([], schema)
        self.assertTrue(config_suite.valid)

        for req in (True, False):
            schema[MK.Required] = req
            with self.assertRaises(ValueError) as err:
                configsuite.ConfigSuite([], schema)
            self.assertIn("Required can only be used for BasicType", str(err.exception))

    def test_named_dict_cannot_spec_required(self):
        schema = {
            MK.Type: configsuite.types.NamedDict,
            MK.Content: {"key": {MK.Type: configsuite.types.Integer}},
        }
        config_suite = configsuite.ConfigSuite({"key": 42}, schema)
        self.assertTrue(config_suite.valid)

        for req in (True, False):
            schema[MK.Required] = req
            with self.assertRaises(ValueError) as err:
                configsuite.ConfigSuite({"key": 42}, schema)
            self.assertIn("Required can only be used for BasicType", str(err.exception))

    def test_dict_cannot_spec_required(self):
        schema = {
            MK.Type: configsuite.types.Dict,
            MK.Content: {
                MK.Key: {MK.Type: configsuite.types.String},
                MK.Value: {MK.Type: configsuite.types.String},
            },
        }
        config_suite = configsuite.ConfigSuite({}, schema)
        self.assertTrue(config_suite.valid)

        for req in (True, False):
            schema[MK.Required] = req
            with self.assertRaises(ValueError) as err:
                configsuite.ConfigSuite({}, schema)
            self.assertIn("Required can only be used for BasicType", str(err.exception))

    def test_date(self):
        timestamp = datetime.datetime(2015, 1, 2, 3, 4, 5)
        date = datetime.date(2010, 1, 1)
        types = [timestamp, date, "fdnsjk", "", 0, -10, [], 1, 1.2, {}, None]

        for date in types:
            raw_config = data.pets.build_config()
            raw_config["pet"]["birthday"] = date
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            if isinstance(date, datetime.date):
                self.assertTrue(config_suite.valid)
                self.assertEqual(date, config_suite.snapshot.pet.birthday)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(("pet", "birthday"), err.key_path)

    def test_datetime(self):
        timestamp = datetime.datetime(2015, 1, 2, 3, 4, 5)
        date = datetime.date(2010, 1, 1)
        types = [timestamp, date, "fdnsjk", "", 0, -10, [], 1, 1.2, {}, None]

        for date in types:
            raw_config = data.pets.build_config()
            raw_config["pet"]["timestamp"] = date
            config_suite = configsuite.ConfigSuite(raw_config, data.pets.build_schema())

            if isinstance(date, datetime.datetime):
                self.assertTrue(config_suite.valid)
                self.assertEqual(date, config_suite.snapshot.pet.timestamp)
                self.assertEqual(0, len(config_suite.errors))
            else:
                self.assertFalse(config_suite.valid)
                self.assertEqual(1, len(config_suite.errors))
                err = config_suite.errors[0]
                self.assertIsInstance(err, configsuite.InvalidTypeError)
                self.assertEqual(("pet", "timestamp"), err.key_path)

    def test_booleanresult_equality(self):
        for x, assert_func in zip((True, False), (self.assertTrue, self.assertFalse)):
            res = configsuite.types.BooleanResult(x, "message", "input")
            assert_func(res)
            msg = "message is {} on input 'input'".format(str(x).lower())
            self.assertEqual(res.msg, msg)

            res_of_res = configsuite.types.BooleanResult(res, "message", "input")
            assert_func(res_of_res)
            msg = "message is {} on input 'input'".format(str(x).lower())
            self.assertEqual(res_of_res.msg, msg)

    def test_invalid_boolean_result_type(self):
        elements = [1, 1.2, "dsa", {"test": 2}, [1, 2, 3], ("2", "3")]
        for elem in elements:
            with self.assertRaises(TypeError):
                configsuite.types.BooleanResult(elem, "incorrect", "type")



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import collections
import copy
import jinja2

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


_VARIABLES = {"color": "blue", "secret_number": 42, "animal": "cow"}
_VAR_START = "<"
_VAR_END = ">"


def _render_variables(variables, jinja_env):
    variables = copy.deepcopy(variables)
    for _ in enumerate(variables):
        rendered_values = []
        for key, value in variables.items():
            try:
                variables[key] = jinja_env.from_string(value).render(variables)
                rendered_values.append(variables[key])
            except TypeError:
                continue

    if any([_VAR_START in val for val in rendered_values]):
        raise ValueError("Circular dependencies")

    return variables


def _render(template, definitions=None):
    if definitions is None:
        definitions = {}

    variables = copy.deepcopy(_VARIABLES)
    variables.update(definitions)
    jinja_env = jinja2.Environment(
        variable_start_string=_VAR_START, variable_end_string=_VAR_END, autoescape=True
    )

    try:
        variables = _render_variables(variables, jinja_env)
        jinja_template = jinja_env.from_string(template)
        return jinja_template.render(variables)
    except TypeError:
        return template


@configsuite.transformation_msg("Renders Jinja template using definitions")
def _context_render(elem, context):
    return _render(elem, definitions=context.definitions)


@configsuite.transformation_msg("Renders Jinja template")
def _render_no_vars(elem):
    return _render(elem)


def build_schema_no_definitions():
    return {
        MK.Type: types.List,
        MK.Content: {
            MK.Item: {MK.Type: types.String, MK.Transformation: _render_no_vars}
        },
    }


def build_config_no_definitions():
    return [
        "This is a story about a <animal>.",
        "It had a <color> house.",
        "And the password to enter was <secret_number>.",
        "The end.",
    ]


def template_results_no_definitions():
    return (
        "This is a story about a cow.",
        "It had a blue house.",
        "And the password to enter was 42.",
        "The end.",
    )


def extract_templating_context(configuration):
    Context = collections.namedtuple("TemplatingContext", ["definitions"])
    return Context(definitions=copy.deepcopy(configuration.definitions))


def build_schema_with_definitions():
    return {
        MK.Type: types.NamedDict,
        MK.Content: {
            "definitions": {
                MK.Type: types.Dict,
                MK.Content: {
                    MK.Key: {MK.Type: types.String},
                    MK.Value: {MK.Type: types.String},
                },
            },
            "templates": {
                MK.Type: types.List,
                MK.Content: {
                    MK.Item: {
                        MK.Type: types.String,
                        MK.ContextTransformation: _context_render,
                    }
                },
            },
        },
    }


def build_config_with_definitions():
    return {
        "definitions": {"animal": "pig", "habitants": "<animal>, cow and monkey"},
        "templates": [
            "This is a story about a <animal>.",
            "It had a <color> house.",
            "And the password to enter was <secret_number>.",
            "If you entered the house you would meet: <habitants>",
            "The end.",
        ],
    }


def template_results_with_definitions():
    return (
        "This is a story about a pig.",
        "It had a blue house.",
        "And the password to enter was 42.",
        "If you entered the house you would meet: pig, cow and monkey",
        "The end.",
    )



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest
import configsuite


class TestElemValidatorMsg(unittest.TestCase):
    def test_true(self):
        @configsuite.validator_msg("This should always be true")
        def tautology():
            return True

        self.assertTrue(tautology())

    def test_false(self):
        @configsuite.validator_msg("This should always be false")
        def self_contradiction():
            return False

        self.assertFalse(self_contradiction())

    def test_func_msg(self):
        msg = "This function is self-imploding"

        @configsuite.validator_msg(msg)
        def implode(_):
            raise Exception("Thou shalt not call me")

        self.assertEqual(msg, implode.msg)

    def test_result_msg(self):
        msg = "The indentify function"

        @configsuite.validator_msg(msg)
        def identity(arg):
            return arg

        res_msg_fmt = "{} is {} on input '{}'"
        for x in (True, False):
            expected_msg = res_msg_fmt.format(msg, str(x).lower(), x)
            # pylint: disable=no-member
            self.assertEqual(expected_msg, identity(x).msg)

    def test_multi_arg_msg(self):
        msg = "The and function"

        @configsuite.validator_msg(msg)
        def _and(arg1, arg2):
            return arg1 and arg2

        expected_msg = "The and function is false on input 'True, False'"
        # pylint: disable=no-member
        self.assertEqual(expected_msg, _and(True, False).msg)

    def test_result_msg_default_keyword(self):
        msg = "The indentify function"

        @configsuite.validator_msg(msg)
        def identity(arg, return_none=False):
            return None if return_none else arg

        res_msg_fmt = "{} is {} on input '{}'"
        for x in (True, False):
            expected_msg = res_msg_fmt.format(msg, str(x).lower(), x)
            # pylint: disable=no-member
            self.assertEqual(expected_msg, identity(x).msg)

    def test_keyword_arg(self):
        msg = "The valid sum function"

        @configsuite.validator_msg(msg)
        def valid_sum(a, b, target_sum=42):
            return a + b == target_sum

        expected_msg = msg + " is true on input '8, 12, target_sum=20'"
        self.assertEqual(expected_msg, valid_sum(8, 12, target_sum=20).msg)



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import collections
import copy
import unittest
import sys

import configsuite
from configsuite import MetaKeys as MK
from .data import templating


class TestContextTransformations(unittest.TestCase):
    def test_context_transformation_valid(self):
        raw_config = templating.build_config_with_definitions()
        suite = configsuite.ConfigSuite(
            raw_config,
            templating.build_schema_with_definitions(),
            extract_transformation_context=templating.extract_templating_context,
        )
        self.assertTrue(suite.valid, suite.errors)
        self.assertEqual(
            templating.template_results_with_definitions(), suite.snapshot.templates
        )

    def test_context_transformation_layers(self):
        raw_config = templating.build_config_with_definitions()
        top_layer = {"definitions": {"animal": "horse"}}
        suite = configsuite.ConfigSuite(
            top_layer,
            templating.build_schema_with_definitions(),
            layers=(raw_config,),
            extract_transformation_context=templating.extract_templating_context,
        )
        self.assertTrue(suite.valid, suite.errors)
        self.assertEqual(
            tuple(
                [
                    line.replace("pig", "horse")
                    for line in templating.template_results_with_definitions()
                ]
            ),
            suite.snapshot.templates,
        )

    def test_context_transformation_push(self):
        raw_config = templating.build_config_with_definitions()
        top_layer = {"definitions": {"animal": "horse"}}
        suite = configsuite.ConfigSuite(
            raw_config,
            templating.build_schema_with_definitions(),
            extract_transformation_context=templating.extract_templating_context,
        )

        layered_suite = suite.push(top_layer)
        self.assertTrue(layered_suite.valid, layered_suite.errors)
        self.assertEqual(
            tuple(
                [
                    line.replace("pig", "horse")
                    for line in templating.template_results_with_definitions()
                ]
            ),
            layered_suite.snapshot.templates,
        )

    def test_context_transformation_exception(self):
        raw_config = templating.build_config_with_definitions()

        schema = templating.build_schema_with_definitions()
        temp_schema = schema[MK.Content]["templates"][MK.Content][MK.Item]
        _render = temp_schema[MK.ContextTransformation]
        transformation_msg = "Renders Jinja template until the end"
        error_msg = "Did not expect to reach the end!"

        @configsuite.transformation_msg(transformation_msg)
        def _no_end_render(elem, context):
            if elem == "The end.":
                raise ValueError(error_msg)
            return _render(elem, context)

        temp_schema[MK.ContextTransformation] = _no_end_render

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=templating.extract_templating_context,
        )
        self.assertFalse(suite.valid)

        self.assertEqual(1, len(suite.errors))
        err = suite.errors[0]
        self.assertIsInstance(err, configsuite.TransformationError)
        error_fmt = "'{}' failed on input '{}' with error '{}'"
        expected_error = error_fmt.format(transformation_msg, "The end.", error_msg)
        self.assertEqual(expected_error, err.msg)
        self.assertEqual(("templates", 4), err.key_path)

    def test_context_transformation_context_extraction_exception(self):
        def _extract_context_with_errors(configuration):
            if ("raise", "exception") in configuration.definitions:
                raise Exception("I'm containing errors")
            return templating.extract_templating_context(configuration)

        raw_config = templating.build_config_with_definitions()
        raw_config["definitions"]["raise"] = "exception"

        suite = configsuite.ConfigSuite(
            raw_config,
            templating.build_schema_with_definitions(),
            extract_transformation_context=_extract_context_with_errors,
        )

        self.assertFalse(suite.valid)
        self.assertEqual(1, len(suite.errors))
        err = suite.errors[0]
        self.assertIsInstance(err, configsuite.ContextExtractionError)

    def test_context_transformation_list(self):
        def extract_prefix_and_definitions(configuration):
            def_context = templating.extract_templating_context(configuration)

            prefix = "(default_prefix)"
            for key, val in configuration.definitions:
                if key == "prefix":
                    prefix = val

            Context = collections.namedtuple("Context", ["definitions", "prefix"])
            return Context(definitions=def_context.definitions, prefix=prefix)

        raw_config = templating.build_config_with_definitions()
        prefix = "[(unreal prefix)] "
        raw_config["definitions"]["prefix"] = prefix

        @configsuite.transformation_msg("Adds prefix to all strings in list")
        def _context_prefixer(templates, context):
            return tuple([context.prefix + template for template in templates])

        schema = templating.build_schema_with_definitions()
        schema[MK.Content]["templates"][MK.ContextTransformation] = _context_prefixer

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=extract_prefix_and_definitions,
        )

        self.assertTrue(suite.valid)
        self.assertEqual(
            tuple(
                [
                    prefix + template
                    for template in templating.template_results_with_definitions()
                ]
            ),
            suite.snapshot.templates,
        )

    def test_context_transformation_nameddict(self):
        raw_config = templating.build_config_with_definitions()

        def extract_illegal_and_definitions(configuration):
            def_context = templating.extract_templating_context(configuration)

            Context = collections.namedtuple("Context", ["definitions", "inject"])
            return Context(definitions=def_context.definitions, inject=True)

        @configsuite.transformation_msg("Injects key `illegal` mapping to `True`")
        def _inject_illegal(elem, context):
            elem = copy.deepcopy(elem)
            if context.inject:
                elem["inject"] = True
            return elem

        schema = templating.build_schema_with_definitions()
        schema[MK.ContextTransformation] = _inject_illegal

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=extract_illegal_and_definitions,
        )

        self.assertFalse(suite.valid)
        self.assertEqual(1, len(suite.errors))
        self.assertIsInstance(suite.errors[0], configsuite.UnknownKeyError)

    def test_context_transformation_dict(self):
        raw_config = templating.build_config_with_definitions()

        def extract_templates_and_definitions(configuration):
            def_context = templating.extract_templating_context(configuration)

            Context = collections.namedtuple("Context", ["definitions", "templates"])
            return Context(
                definitions=def_context.definitions, templates=configuration.templates
            )

        @configsuite.transformation_msg("Injects all templates as definitions")
        def template_injector(definitions, context):
            definitions = copy.deepcopy(definitions)
            for idx, template in enumerate(context.templates):
                definitions["template[{}]".format(idx)] = template
            return definitions

        schema = templating.build_schema_with_definitions()
        schema[MK.Content]["definitions"][MK.ContextTransformation] = template_injector

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=extract_templates_and_definitions,
        )

        self.assertTrue(suite.valid)
        self.assertEqual(
            sum(map(len, raw_config.values())), len(suite.snapshot.definitions)
        )

    def test_context_transformation_non_readable_results(self):
        raw_config = templating.build_config_with_definitions()

        @configsuite.transformation_msg("Deforming elements")
        def _deformer(elem, _):
            return [elem]

        schema = templating.build_schema_with_definitions()
        schema[MK.ContextTransformation] = _deformer

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=templating.extract_templating_context,
        )
        self.assertFalse(suite.readable, suite.errors)
        self.assertEqual(1, len(suite.errors))

    def test_context_transformation_non_readable_transformation_result(self):
        raw_config = templating.build_config_with_definitions()

        @configsuite.transformation_msg("Deforming elements")
        def _deformer(elem):
            return [elem]

        schema = templating.build_schema_with_definitions()
        schema[MK.Transformation] = _deformer
        schema[MK.ContextTransformation] = lambda *_: sys.exit(1)

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=templating.extract_templating_context,
        )
        self.assertFalse(suite.readable, suite.errors)
        self.assertEqual(1, len(suite.errors))
        self.assertIsInstance(suite.errors[0], configsuite.InvalidTypeError)

    def test_context_transformation_after_transformation(self):
        raw_config = templating.build_config_with_definitions()
        prefix = "[(unreal prefix)] "

        @configsuite.transformation_msg("Adds prefix to a string")
        def _add_prefix(template):
            return prefix + template

        schema = templating.build_schema_with_definitions()
        template_schema = schema[MK.Content]["templates"][MK.Content][MK.Item]
        template_schema[MK.Transformation] = _add_prefix

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=templating.extract_templating_context,
        )

        self.assertTrue(suite.valid)
        self.assertEqual(
            tuple(
                [
                    prefix + template
                    for template in templating.template_results_with_definitions()
                ]
            ),
            suite.snapshot.templates,
        )

    def test_context_transformation_prior_to_validation(self):
        raw_config = templating.build_config_with_definitions()
        raw_config["templates"].append(">")

        @configsuite.validator_msg("Are all template traces gone")
        def _no_template_trace(template):
            return "<" not in template and ">" not in template

        schema = templating.build_schema_with_definitions()
        template_schema = schema[MK.Content]["templates"][MK.Content][MK.Item]
        template_schema[MK.ElementValidators] = (_no_template_trace,)

        suite = configsuite.ConfigSuite(
            raw_config,
            schema,
            extract_transformation_context=templating.extract_templating_context,
        )

        self.assertFalse(suite.valid)
        self.assertEqual(1, len(suite.errors))
        self.assertIsInstance(suite.errors[0], configsuite.InvalidValueError)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest

import configsuite
from configsuite import MetaKeys as MK

from . import data


class TestDict(unittest.TestCase):
    def test_valid_store_config(self):
        raw_config = data.store.build_config()
        store_schema = data.store.build_schema()
        config_suite = configsuite.ConfigSuite(raw_config, store_schema)

        self.assertTrue(config_suite.valid)
        prices = [(elem.key, elem.value) for elem in config_suite.snapshot.prices]
        self.assertEqual(
            sorted(tuple(raw_config["prices"].items())), sorted(tuple(prices))
        )

    def test_invalid_schema_no_key(self):
        raw_config = data.store.build_config()
        store_schema = data.store.build_schema()
        store_schema[MK.Content]["prices"][MK.Content].pop(MK.Key)

        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, store_schema)

    def test_invalid_schema_no_value(self):
        raw_config = data.store.build_config()
        store_schema = data.store.build_schema()
        store_schema[MK.Content]["prices"][MK.Content].pop(MK.Value)

        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, store_schema)

    def test_invalid_schema_extra_spec(self):
        raw_config = data.store.build_config()
        store_schema = data.store.build_schema()
        store_schema[MK.Content]["prices"][MK.Content]["monkey"] = "patch"

        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, store_schema)

    def test_valid_store_config_faulty_key(self):
        raw_config = data.store.build_config()
        raw_config["prices"][123] = 1
        store_schema = data.store.build_schema()
        config_suite = configsuite.ConfigSuite(raw_config, store_schema)

        self.assertFalse(config_suite.valid)
        self.assertEqual(1, len(config_suite.errors))
        err = config_suite.errors[0]
        self.assertIsInstance(err, configsuite.InvalidTypeError)

    def test_valid_store_config_faulty_value(self):
        raw_config = data.store.build_config()
        raw_config["prices"]["pig"] = "very expensive"
        store_schema = data.store.build_schema()
        config_suite = configsuite.ConfigSuite(raw_config, store_schema)

        self.assertFalse(config_suite.valid)
        self.assertEqual(1, len(config_suite.errors))
        err = config_suite.errors[0]
        self.assertIsInstance(err, configsuite.InvalidTypeError)

    def test_valid_advanded_store_config(self):
        raw_config = data.advanced_store.build_config()
        store_schema = data.advanced_store.build_schema()
        config_suite = configsuite.ConfigSuite(raw_config, store_schema)

        self.assertTrue(config_suite.valid)
        # Map all not listed values to None
        raw_prices = [
            (key, {key: value.get(key) for key in ("unit", "kilo", "gram")})
            for (key, value) in raw_config["prices"].items()
        ]

        # Extract and format prices from config
        prices = [
            (
                elem.key,
                {
                    "unit": elem.value.unit,
                    "kilo": elem.value.kilo,
                    "gram": elem.value.gram,
                },
            )
            for elem in config_suite.snapshot.prices
        ]
        self.assertEqual(sorted(raw_prices), sorted(prices))

    def test_advanced_store_config_faulty_schema(self):
        raw_config = data.advanced_store.build_config()
        store_schema = data.advanced_store.build_schema()
        store_schema[MK.Content]["prices"][MK.Content][MK.Value][MK.Content][
            "kilo"
        ].pop(MK.Type)

        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, store_schema)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""

import os
import importlib

# The getargspec/getfullargspec should be dropped
# for the "signature" module when python2 is not supported
try:
    from inspect import getfullargspec as inspect_method
except ImportError:
    from inspect import getargspec as inspect_method

from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged_required
from docutils import nodes

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types
from configsuite.extension.ext_nodes import (
    visit_children,
    visit_container,
    visit_content,
    visit_header,
    depart_children,
    depart_container,
    depart_content,
    depart_header,
    cs_children,
    cs_container,
    cs_content,
    cs_header,
)

BASIC_EXAMPLE = {
    "string": "default",
    "integer": "0",
    "number": "0.0",
    "bool": "True",
    "date": "01.01.1999",
    "datetime": "01.01.1999",
}


def scb_static_path(app):
    static_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "_static"))
    app.config.html_static_path.append(static_path)


def generate_section_content(schema, title):
    section_content = cs_content(
        classes=["cs_content"], ids=["cs_{}_content".format(title)]
    )

    section_content += nodes.paragraph(text=schema.get(MK.Description, ""))

    for config_key in [MK.ElementValidators, MK.ContextValidators]:
        element_nodes = [
            nodes.paragraph(text=validator.msg)
            for validator in schema.get(config_key, [])
        ]
        section_content.extend(element_nodes)

    for config_key in [
        MK.LayerTransformation,
        MK.ContextTransformation,
        MK.Transformation,
    ]:
        if config_key in schema:
            section_content += nodes.paragraph(text=schema.get(config_key).msg)

    return section_content


def generate_section_children(schema, title):
    section_children = cs_children(
        classes=["cs_children"], ids=["cs_{}_children".format(title)]
    )

    if schema[MK.Type] == types.NamedDict:
        example = dict()
        for key, value in schema[MK.Content].items():
            generate(value, section_children, key)

        for child in section_children.children:
            example[child["parent"]] = child["example"]

    elif schema[MK.Type] == types.List:
        generate(
            schema[MK.Content][MK.Item], section_children, parent_title="List Item"
        )
        example = [section_children.children[0]["example"]]

    elif schema[MK.Type] == types.Dict:
        example = dict()
        generate(schema[MK.Content][MK.Key], section_children, parent_title="Key")
        generate(schema[MK.Content][MK.Value], section_children, parent_title="Value")
        example[section_children.children[0]["example"]] = section_children.children[1][
            "example"
        ]

    elif isinstance(schema[MK.Type], types.BasicType):
        example = "{}".format(BASIC_EXAMPLE.get(schema[MK.Type].name, "undefined"))
        section_children += nodes.paragraph(
            text=":type: {_type}".format(_type=schema[MK.Type].name)
        )

    else:
        err_msg = "Unexpected type ({}) in schema while generating documentation."
        raise TypeError(err_msg.format(schema[MK.Type]))

    return section_children


def generate(schema, parent_section, parent_title=""):
    title = schema.get(MK.Description, "default_title")[:15]
    title = title.lower()
    title = title.replace(" ", "_")
    example = None

    section = cs_container(
        classes=["cs_container"], ids=["cs_{}_container".format(title)]
    )

    section += cs_header(classes=["cs_header"], text=parent_title)

    section_content = generate_section_content(schema, title)
    section_children = generate_section_children(schema, title)

    section_content += section_children

    section += section_content
    section["example"] = example
    section["parent"] = parent_title

    parent_section += section


class ConfigsuiteDocDirective(Directive):
    has_content = True
    option_spec = {"module": unchanged_required}

    def run(self):
        if "module" not in self.options:
            raise self.error("Required :module: argument missing")

        schema_function_string = self.options.get("module")

        if "." not in schema_function_string:
            raise self.error(
                "Configsuite :module: has invalid format. "
                + "Expected 'module.function', got {}".format(schema_function_string)
            )

        module_str, func_str = schema_function_string.rsplit(".", 1)

        try:
            module = importlib.import_module(module_str)
        except (ImportError) as e:
            raise self.error(str(e))

        try:
            schema_func = getattr(module, func_str)
        except (AttributeError) as e:
            raise self.error(str(e))

        if not callable(schema_func):
            err_msg = (
                "The module.function given at Configsuite :module: "
                + "must provide a callable function, the '{func_name}' was not"
            )
            raise self.error(err_msg.format(func_name=func_str))

        insp = inspect_method(schema_func)  # pylint: disable=deprecated-method
        if len(insp.args) != 0:
            err_msg = (
                "The module.function given at Configsuite :module: "
                + "can not require any arguments, the '{func_name}' "
                + "takes {num_args} argument(s)"
            )
            raise self.error(
                err_msg.format(func_name=func_str, num_args=len(insp.args))
            )

        schema = schema_func()
        if not isinstance(schema, dict):
            err_msg = (
                "The module.function given at Configsuite :module: "
                + "is expected to return a dictionary object\n"
                + "The function '{func_name}' returned a {return_type}"
            )
            raise self.error(
                err_msg.format(func_name=func_str, return_type=type(schema))
            )

        section = cs_container(classes=["cs_top_container"])
        generate(schema, section)
        return [section]


def setup(app):
    # Add our static path
    app.connect("builder-inited", scb_static_path)

    # Add relevant code to headers
    app.add_css_file("configsuite_ext.css")
    app.add_js_file("configsuite_ext.js")

    app.add_node(
        cs_container, html=(visit_container, depart_container),
    )

    app.add_node(
        cs_header, html=(visit_header, depart_header),
    )

    app.add_node(
        cs_content, html=(visit_content, depart_content),
    )

    app.add_node(
        cs_children, html=(visit_children, depart_children),
    )

    app.add_directive("configsuite", ConfigsuiteDocDirective)

    return {"version": configsuite.__version__}



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import collections

import configsuite
from configsuite import MetaKeys as MK


TransformationResult = collections.namedtuple(
    "TransformationResult", ("success", "errors", "result")
)


class Transformer(object):
    def __init__(
        self, schema, transformation_type, transformation_context, bottom_up=True
    ):
        self._schema = schema
        self._transformation_type = transformation_type
        self._transformation_context = transformation_context
        self._bottom_up = bottom_up
        self._errors = None

        self._debug = transformation_type == MK.ContextTransformation
        self._debug &= transformation_context is not None

    def transform(self, config):
        self._errors = []
        transformed_config = self._transform(config, self._schema, ())

        return TransformationResult(
            success=len(self._errors) == 0,
            errors=tuple(self._errors),
            result=transformed_config,
        )

    def _transform(self, config, schema, key_path):
        if not self._bottom_up:
            config = self._apply_single_transformation(config, schema, key_path)

        data_type = schema[MK.Type]
        if isinstance(data_type, configsuite.types.BasicType):
            pass
        elif data_type == configsuite.types.List:
            config = self._transform_list(config, schema, key_path)
        elif data_type == configsuite.types.NamedDict:
            config = self._transform_named_dict(config, schema, key_path)
        elif data_type == configsuite.types.Dict:
            config = self._transform_dict(config, schema, key_path)
        else:
            msg = "Encountered unknown type {} while building raw config"
            raise TypeError(msg.format(str(data_type)))

        if self._bottom_up:
            config = self._apply_single_transformation(config, schema, key_path)

        return config

    def _transform_list(self, config, schema, key_path):
        if not configsuite.types.List.validate(config):
            return config

        item_schema = schema[MK.Content][MK.Item]
        return tuple(
            [
                self._transform(item, item_schema, key_path + (idx,))
                for idx, item in enumerate(config)
            ]
        )

    def _transform_named_dict(self, config, schema, key_path):
        if not configsuite.types.NamedDict.validate(config):
            return config

        content_schema = schema[MK.Content]
        transformed_config = {}
        for key, value in config.items():
            if key not in content_schema:
                transformed_config[key] = value
                continue

            transformed_config[key] = self._transform(
                value, content_schema[key], key_path + (key,)
            )

        return transformed_config

    def _transform_dict(self, config, schema, key_path):
        if not configsuite.types.Dict.validate(config):
            return config

        key_schema = schema[MK.Content][MK.Key]
        value_schema = schema[MK.Content][MK.Value]

        transformed_config = {}
        for key, value in config.items():
            tkey = self._transform(key, key_schema, key_path + (key,))
            tval = self._transform(value, value_schema, key_path + (key,))
            transformed_config[tkey] = tval

        return transformed_config

    def _apply_single_transformation(self, config, schema, key_path):
        if self._transformation_type not in schema:
            return config

        transformation = schema[self._transformation_type]
        try:
            return transformation(config, *(self._transformation_context))
        # pylint: disable=broad-except
        except Exception as e:
            error_fmt = "'{}' failed on input '{}' with error '{}'"
            error_msg = error_fmt.format(transformation.msg, config, str(e))
            self._errors.append(configsuite.TransformationError(error_msg, key_path))

        return config



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import copy
import configsuite
import collections


from .schema import assert_valid_schema
from .meta_keys import MetaKeys as MK


class ConfigSuite(object):
    """A `Suite` exposing the functionality of Config Suite in a unified manner.

    The intended usage of Config Suite is via this immutable suite. It is constructed
    with a `schema` describing the structure of a configuration, together with
    a `raw_config` and possibly additional `layers`.

    Parameters
    ----------
    raw_config
        The configuration taking precedence.
    schema
        A description of the structure of a valid configuration, together with
        actions that are to be carried out.
    layers: iterable of layers, optional
        Additional layers of configuration. A layer takes precedence over all
        other layers following it in the given sequence. Note that `raw_config`
        takes precedence over all elements of `layers`.
    extract_validation_context: callable, optional
        Callable that extracts the context used for validation. The callable is
        given a snapshot of the configuration as argument. Defaults to the
        constant function always returning `None`.
    extract_transformation_context: callable, optional
        Callable that extracts the context used for transformations. The
        callable is given a snapshot of the configuration as argument. Defaults
        to the constant function always returning `None`.
    deduce_required: bool, optional
        Boolean that enables future behaviour of deducing whether a schema
        entry is `required` by inspecting `allow_none` and `default`. In
        particular, using `required` in schemas as well as not setting
        `deduce_required=True` is deprecated.



    Raises
    ------
    TypeError, KeyError, ValueError
        Approperiate errors are raised if provided with an invalid schema. Note
        that errors are independent of `raw_config` and `layers` and hence not
        user input responsive.
    """

    def __init__(
        self,
        raw_config,
        schema,
        layers=(),
        extract_validation_context=lambda snapshot: None,
        extract_transformation_context=lambda snapshot: None,
        deduce_required=False,
    ):
        assert_valid_schema(schema, deduce_required=deduce_required)
        self._layers = tuple(
            [copy.deepcopy(layer) for layer in tuple(layers) + (raw_config,)]
        )
        self._schema = copy.deepcopy(schema)
        self._extract_validation_context = extract_validation_context
        self._extract_transformation_context = extract_transformation_context

        self._readable = True
        self._valid = True
        self._errors = ()
        self._snapshot = None

        self._cached_merged_config = self._build_merged_config()
        if self._readable:
            self._validate_final()
        self._assert_state()

    @property
    def valid(self):
        """A boolean indicating whether the resulting configuration was deemed
        valid according to the schema."""
        return self._valid

    @property
    def errors(self):
        """An iterable of `configsuite.errors` that occurred during validation. Is always
        empty if `valid` is `True` and non-empty otherwise."""
        return self._errors

    @property
    def readable(self):
        """A boolean indicating whether the resulting configuration was deemed
        readable. Is always `True` if `valid` is `True`. For the consequences
        of being readable, see the documentation of `snapshot`."""
        return self._readable

    @property
    def snapshot(self):
        """A complete, immutable representation of the resulting configuration.

        Raises
        ------
        AssertionError
            Raised on access if `suite` is not `readable`.
        """
        if not self.readable:
            err_msg = "Cannot build snapshot of unreadable configuration."
            raise AssertionError(err_msg)

        if self._snapshot is None:
            self._snapshot = self._build_snapshot(self._merged_config, self._schema)

        return self._snapshot

    @property
    def _validation_context(self):
        return self._extract_validation_context(self.snapshot)

    def push(self, raw_config):
        """Builds a new suite with `raw_config` on top of the current suite.

        Builds a new suite with the same schema, but with `raw_config` on top
        of the layers in the current suite.

        Parameters
        ----------
        raw_config:
            A configuration that is to take precedence over all layers in the
            current suite.

        Returns
        -------
        A new `ConfigSuite` with `raw_config` as the first layer.
        """
        return ConfigSuite(
            raw_config,
            self._schema,
            layers=self._layers,
            extract_validation_context=self._extract_validation_context,
            extract_transformation_context=self._extract_transformation_context,
        )

    @property
    def _merged_config(self):
        """The merged config cannot be built and will also not be utilized if the
        config is not readable. Due to this it is built lazily.
        """
        if not self.readable:
            err_msg = (
                "Internal error: "
                "Cannot build merged_config of unreadable configuration."
            )
            raise AssertionError(err_msg)

        if self._cached_merged_config is None:
            err_msg = (
                "Internal error: "
                "Merged config cannot be accessed before cache is built."
            )
            raise AssertionError(err_msg)

        return self._cached_merged_config

    def _build_merged_config(self):
        layers = self._build_transformed_layers()

        self._validate_readability(layers)
        if not self.readable:
            return None

        merged_config = self._build_initial_merged_config(layers, self._schema)
        merged_config = self._apply_transformations(merged_config)

        self._validate_readability((merged_config,))
        if not self.readable:
            return None

        merged_config = self._apply_context_transformations(merged_config)

        self._validate_readability((merged_config,))
        if not self.readable:
            return None

        return merged_config

    def _build_transformed_layers(self):
        layer_transformer = configsuite.Transformer(
            self._schema, MK.LayerTransformation, (), bottom_up=False
        )
        layers = []
        for layer in self._layers:
            trans_layer = layer_transformer.transform(layer)
            self._errors += trans_layer.errors
            layers.append(trans_layer.result)

        self._valid &= len(self._errors) == 0
        return layers

    def _build_initial_named_dict_merged_config(self, layers, schema):
        rec = self._build_initial_merged_config
        content_schema = schema[MK.Content]

        def is_collection(schema_type):
            return isinstance(schema_type, configsuite.types.Collection)

        def is_defaultable(schema_item):
            return MK.Default in schema_item or is_collection(schema_item[MK.Type])

        layer_keys = set((key for layer in layers for key in layer.keys()))
        defaultable_keys = set(
            key for key, val in content_schema.items() if is_defaultable(val)
        )

        config = {}
        for key in layer_keys | defaultable_keys:
            child_layers = tuple([layer[key] for layer in layers if key in layer])

            if key in defaultable_keys:
                key_type = content_schema[key][MK.Type]
                if is_collection(key_type):
                    default = key_type.create_empty()
                else:
                    default = content_schema[key].get(MK.Default)
                child_layers = (default,) + child_layers

            if len(child_layers) == 0:
                continue

            if key in content_schema:
                config[key] = rec(child_layers, content_schema[key])
            else:
                config[key] = child_layers[-1]

        return config

    def _build_initial_dict_merged_config(self, layers, schema):
        rec = self._build_initial_merged_config
        content_schema = schema[MK.Content]
        config = {}
        all_keys = (key for layer in layers for key in layer.keys())
        for key in all_keys:
            child_layers = tuple([layer[key] for layer in layers if key in layer])
            if len(child_layers) == 0:
                continue
            config[key] = rec(child_layers, content_schema[MK.Value])
        return config

    def _build_initial_list_merged_config(self, layers, schema):
        rec = self._build_initial_merged_config
        item_schema = schema[MK.Content][MK.Item]
        config = []
        for layer in layers:
            config += [rec((item,), item_schema) for item in layer]
        return tuple(config)

    def _build_initial_merged_config(self, layers, schema):
        data_type = schema[MK.Type]

        if isinstance(data_type, configsuite.types.BasicType):
            return layers[-1]
        elif data_type == configsuite.types.List:
            return self._build_initial_list_merged_config(layers, schema)
        elif data_type == configsuite.types.NamedDict:
            return self._build_initial_named_dict_merged_config(layers, schema)
        elif data_type == configsuite.types.Dict:
            return self._build_initial_dict_merged_config(layers, schema)
        else:
            msg = "Encountered unknown type {} while building raw config"
            raise TypeError(msg.format(str(data_type)))

    def _apply_transformations(self, config):
        transformer = configsuite.Transformer(self._schema, MK.Transformation, ())
        trans_res = transformer.transform(config)
        self._errors += trans_res.errors
        self._valid &= len(trans_res.errors) == 0
        return trans_res.result

    def _apply_context_transformations(self, config):
        prelim_snapshot = self._build_snapshot(config, self._schema)
        try:
            context = self._extract_transformation_context(prelim_snapshot)
        # pylint: disable=broad-except
        except Exception as e:
            self._valid = False
            self._errors += (configsuite.ContextExtractionError(str(e), ()),)
            return config

        context_transformer = configsuite.Transformer(
            self._schema, MK.ContextTransformation, (context,)
        )
        trans_res = context_transformer.transform(config)
        self._errors += trans_res.errors
        self._valid &= len(trans_res.errors) == 0
        return trans_res.result

    def _build_named_dict_snapshot(self, config, schema):
        data_type = schema[MK.Type]
        content_schema = schema[MK.Content]
        dict_name = data_type.name
        dict_keys = sorted(content_schema.keys())
        dict_collection = collections.namedtuple(dict_name, dict_keys)

        return dict_collection(
            **{
                key: self._build_snapshot(config.get(key), content_schema[key])
                for key in content_schema
            }
        )

    def _build_list_snapshot(self, config, schema):
        item_schema = schema[MK.Content][MK.Item]
        return tuple([self._build_snapshot(elem, item_schema) for elem in config])

    def _build_dict_snapshot(self, config, schema):
        key_schema = schema[MK.Content][MK.Key]
        value_schema = schema[MK.Content][MK.Value]

        Pair = collections.namedtuple("KeyValuePair", ["key", "value"])
        return tuple(
            [
                Pair(
                    self._build_snapshot(key, key_schema),
                    self._build_snapshot(value, value_schema),
                )
                for key, value in config.items()
            ]
        )

    def _build_snapshot(self, config, schema):
        if config is None:
            return None

        data_type = schema[MK.Type]
        if isinstance(data_type, configsuite.BasicType):
            return config
        elif data_type == configsuite.types.NamedDict:
            return self._build_named_dict_snapshot(config, schema)
        elif data_type == configsuite.types.List:
            return self._build_list_snapshot(config, schema)
        elif data_type == configsuite.types.Dict:
            return self._build_dict_snapshot(config, schema)
        else:
            msg = "Encountered unknown type {} while building snapshot"
            raise TypeError(msg.format(str(data_type)))

    def _validate_readability(self, layers):
        readable_errors = (configsuite.UnknownKeyError, configsuite.MissingKeyError)

        def _not_container(schema):
            return not isinstance(schema[MK.Type], configsuite.types.Collection)

        container_validator = configsuite.Validator(
            self._schema, stop_condition=_not_container, apply_validators=False
        )
        container_errors = []
        for idx, layer in enumerate(layers):
            val_res = container_validator.validate(layer)
            container_errors += [
                error.create_layer_error(idx)
                for error in val_res.errors
                if not isinstance(error, readable_errors)
            ]

        self._readable &= len(container_errors) == 0
        self._valid &= self._readable
        self._errors += tuple(container_errors)

    def _validate_final(self):
        if not self.readable:
            err_msg = "Internal error: Not readable when doing final validation"
            raise AssertionError(err_msg)
        self._assert_state()

        validator = configsuite.Validator(self._schema)
        val_res = validator.validate(self._merged_config, self._validation_context)
        self._valid &= val_res.valid
        self._errors += val_res.errors

    def _assert_state(self):
        """Asserts that the internal state is consistent. In particular we will
        verify that:
         - not readable => not valid
         - valid <=> no errors
        """
        if not self.readable and self.valid:
            err_msg = "Internal error: Config is valid, but not readable"
            raise AssertionError(err_msg)
        if self.valid and len(self.errors) > 0:
            err_msg = "Internal error: Config is valid, but has errors"
            raise AssertionError(err_msg)
        if not self.valid and len(self.errors) == 0:
            err_msg = "Internal error: Config is not valid, but has no errors"
            raise AssertionError(err_msg)



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""

import unittest
import datetime

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types
from .data import car


class TestDefaultValues(unittest.TestCase):
    def test_untouched_basic_types(self):
        raw_config = car.build_all_default_config()
        car_schema = car.build_schema()
        car_schema[MK.Content]["tire"][MK.Content]["rim"].pop(MK.Default)

        config_suite = configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(config_suite.valid)

        self.assertEqual(17, config_suite.snapshot.tire.dimension)
        self.assertEqual(None, config_suite.snapshot.tire.rim)

    def test_empty_named_dict_default_values(self):
        raw_config = car.build_all_default_config()
        car_schema = car.build_schema()
        config_suite = configsuite.ConfigSuite(raw_config, car_schema)

        self.assertTrue(config_suite.valid)

        self.assertEqual(17, config_suite.snapshot.tire.dimension)
        self.assertEqual("green", config_suite.snapshot.tire.rim)

    def test_default_values(self):
        raw_config = car.build_config()
        car_schema = car.build_schema()
        config_suite = configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(config_suite.valid)

        # Top level basic type within named dict
        self.assertEqual("Norway", config_suite.snapshot.country)

        # Named dict within named dict
        self.assertEqual(15, config_suite.snapshot.tire.dimension)
        self.assertEqual("green", config_suite.snapshot.tire.rim)

        # Named dict within dict
        owner_data = (("Johan", "Svalbard"), ("Svein", "Earth"))
        self.assertEqual(len(owner_data), len(config_suite.snapshot.owner))
        for (expected_name, expected_location), config_owner in zip(
            owner_data, sorted(config_suite.snapshot.owner)
        ):
            self.assertEqual(expected_name, config_owner.value.name)
            self.assertEqual(expected_location, config_owner.value.location)

        # Named dict within list
        incidents_data = (("whereabouts", 0), ("somewhere else", 1))
        self.assertEqual(len(incidents_data), len(config_suite.snapshot.incidents))
        for (expected_location, expected_casualties), config_owner in zip(
            incidents_data, sorted(config_suite.snapshot.incidents)
        ):
            self.assertEqual(expected_location, config_owner.location)
            self.assertEqual(expected_casualties, config_owner.casualties)

    def test_element_transformations_applied(self):
        raw_config = car.build_all_default_config()
        car_schema = car.build_schema()
        tire_schema = car_schema[MK.Content]["tire"]
        tire_schema[MK.Content]["dimension"][MK.Transformation] = car.inch_to_cm
        config_suite = configsuite.ConfigSuite(raw_config, car_schema)

        self.assertTrue(config_suite.valid)
        self.assertAlmostEqual(17 * 2.54, config_suite.snapshot.tire.dimension)

    def test_context_transformation_applied(self):
        raw_config = car.build_all_default_config()
        car_schema = car.build_schema()
        tire_schema = car_schema[MK.Content]["tire"]
        tire_schema[MK.Content]["dimension"][
            MK.ContextTransformation
        ] = car.inch_to_cm_context_based

        config_suite = configsuite.ConfigSuite(
            raw_config, car_schema, extract_validation_context=lambda snapshot: None
        )

        self.assertTrue(config_suite.valid)
        self.assertAlmostEqual(17 * 2.54, config_suite.snapshot.tire.dimension)

    def test_element_validators_applied(self):
        raw_config = car.build_all_default_config()
        car_schema = car.build_schema()
        tire_content = car_schema[MK.Content]["tire"][MK.Content]
        tire_content["dimension"][MK.ElementValidators] = [car.is_valid_dimension]

        config_suite = configsuite.ConfigSuite(raw_config, car_schema)

        self.assertFalse(config_suite.valid)
        for err in config_suite.errors:
            self.assertTrue(
                isinstance(err, configsuite.validation_errors.InvalidValueError)
            )
            self.assertTrue(err.msg.startswith("Is x a valid dimension"))

    def test_context_validators_applied(self):
        raw_config = car.build_config()
        car_schema = car.build_schema()

        date_content = {
            MK.Type: types.DateTime,
            MK.Default: datetime.datetime(1999, 1, 1),
            MK.ContextValidators: [car.is_valid_date],
            MK.Required: False,
            MK.AllowNone: True,
        }

        owner_schema = car_schema[MK.Content]["owner"]
        owner_schema[MK.Content][MK.Value][MK.Content]["date"] = date_content

        config_suite = configsuite.ConfigSuite(
            raw_config,
            car_schema,
            extract_validation_context=car.extract_production_date,
        )

        self.assertFalse(config_suite.valid)
        for err in config_suite.errors:
            self.assertTrue(
                isinstance(err, configsuite.validation_errors.InvalidValueError)
            )
            self.assertTrue(err.msg.startswith("Is x a valid date"))

    def test_not_allow_default_for_list(self):
        raw_config = car.build_all_default_config()

        car_schema = car.build_schema()
        schema_value_content = {
            MK.Default: ["default value"],
            MK.Type: types.List,
            MK.Content: {MK.Item: {MK.Type: types.String}},
        }
        car_schema[MK.Content]["incidents"] = schema_value_content

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(
            str(error_context.exception).startswith("Default can only be used")
        )

    def test_not_allow_defaults_for_list_items(self):
        raw_config = car.build_all_default_config()

        car_schema = car.build_schema()
        schema_value_content = {
            MK.Type: types.List,
            MK.Content: {MK.Item: {MK.Type: types.String, MK.Default: "default value"}},
        }
        car_schema[MK.Content]["incidents"] = schema_value_content

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(str(error_context.exception).startswith("Default value is"))

    def test_not_allow_default_for_dict(self):
        raw_config = car.build_all_default_config()

        car_schema = car.build_schema()
        schema_value_content = {
            MK.Default: {"key": 10},
            MK.Type: types.Dict,
            MK.Content: {
                MK.Key: {MK.Type: types.String},
                MK.Value: {MK.Type: types.Number, MK.Default: 10},
            },
        }
        car_schema[MK.Content]["owner"] = schema_value_content

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(
            str(error_context.exception).startswith("Default can only be used")
        )

    def test_not_allow_defaults_for_dict_keys(self):
        raw_config = car.build_all_default_config()

        car_schema = car.build_schema()
        schema_value_content = {
            MK.Type: types.Dict,
            MK.Content: {
                MK.Key: {
                    MK.Type: types.String,
                    MK.Default: "default value",
                    MK.Required: False,
                    MK.AllowNone: True,
                },
                MK.Value: {MK.Type: types.Number},
            },
        }
        car_schema[MK.Content]["owner"] = schema_value_content

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(str(error_context.exception).startswith("Default value is"))

    def test_not_allow_defaults_for_dict_values(self):
        raw_config = car.build_all_default_config()

        car_schema = car.build_schema()
        schema_value_content = {
            MK.Type: types.Dict,
            MK.Content: {
                MK.Key: {MK.Type: types.String},
                MK.Value: {
                    MK.Type: types.Number,
                    MK.Default: 10,
                    MK.Required: False,
                    MK.AllowNone: True,
                },
            },
        }
        car_schema[MK.Content]["owner"] = schema_value_content

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(str(error_context.exception).startswith("Default value is"))

    def test_not_allow_default_for_named_dict(self):
        raw_config = car.build_all_default_config()

        schema_value_content = {
            MK.Type: types.NamedDict,
            MK.Default: {"profile": 200, "width": 20},
            MK.Content: {
                "profile": {MK.Type: types.Number},
                "width": {MK.Type: types.Number},
            },
        }
        car_schema = car.build_schema()
        car_schema[MK.Content]["tire"][MK.Content]["rim"] = schema_value_content

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(raw_config, car_schema)
        self.assertTrue(
            str(error_context.exception).startswith("Default can only be used")
        )



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""

from . import favourite_numbers
from . import store
from . import advanced_store
from . import pets
from . import candy_bag
from . import candidate
from . import hero
from . import transactions
from . import templating
from . import numbers
from . import special_numbers
from . import car



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest
import sys

import configsuite

from . import data


class TestDocGen(unittest.TestCase):
    @unittest.skipIf(sys.version_info < (3, 6), reason="requires python3.6 or higher")
    def test_doc_generate(self):
        """This test is skipped for Python versions prior to 3.6 due to those
        not preserving the order of a dictionary (it is technically not a part
        of the API of 3.6 either, but it is how it is implemented) and hence
        not the output of the doc.generate either.
        """
        schema = data.pets.build_schema()
        docs = configsuite.docs.generate(schema)

        exp_docs = data.pets.build_docs()
        self.assertEqual(exp_docs, docs)

    def test_doc_generate_weak(self):
        """This is a much weaker version of the test_doc_generate. It still
        checks that the generator gives output that could possibly be the same
        as the proper test.
        """
        schema = data.pets.build_schema()
        docs = configsuite.docs.generate(schema)

        exp_docs = data.pets.build_docs()
        self.assertEqual(sorted(exp_docs), sorted(docs))



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import six

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


@configsuite.transformation_msg("Convert ranges and singeltons into list")
def _realize_list(l):
    """If l is not a string, l will be returend. If l is a string it is assumed
    to contain comma separated elements. Each element is assumed to be either
    a singelton or a range. A singleton is a single number, i.e.  7 or 14. A
    range is a lower and upper element of the range separated by a single '-'.
    When provided with a string we will either return a list containing the
    union of all the singeltons and the ranges, or raise a TypeError or
    ValueError if it fails in the process.

    _realize_list('1,2,4-7,14-15') -> [1, 2, 4, 5, 6, 7, 14, 15]
    """
    if not isinstance(l, six.string_types):
        return l

    real_list = []
    for elem in l.split(","):
        bounds = elem.split("-")
        if len(bounds) == 1:
            if "-" in elem:
                raise ValueError("Did not expect '-' in singleton")
            real_list.append(int(elem))
        elif len(bounds) == 2:
            if elem.count("-") != 1:
                raise ValueError("Did expect single '-' in range")
            lower_bound = int(bounds[0])
            upper_bound = int(bounds[1]) + 1

            if lower_bound > upper_bound:
                err_msg = "Lower bound of range expected to be smaller then upper bound"
                raise ValueError(err_msg)

            real_list += range(lower_bound, upper_bound)
        else:
            raise ValueError("Expected at most one '-' in an element")

    return real_list


@configsuite.transformation_msg("Sort and remove duplicates in list")
def _unique_and_sorted(l):
    return tuple(sorted(set(l)))


def build_schema():
    return {
        MK.Type: types.List,
        MK.LayerTransformation: _realize_list,
        MK.Transformation: _unique_and_sorted,
        MK.Content: {MK.Item: {MK.Type: types.Integer}},
    }
